#ifndef shaderdata_h
#define shaderdata_h

#include <cstring>

extern const char *fragment;
extern unsigned int fragment_len;

extern const char *fragment_es;
extern unsigned int fragment_es_len;

extern const char *vertex;
extern unsigned int vertex_len;

extern const char *vertex_es;
extern unsigned int vertex_es_len;

#endif